import json

from cryptojwt.jwk.rsa import import_private_rsa_key_from_file, RSAKey
from cryptojwt.jwe.jwe import factory
from cryptojwt.jwe.jwe_rsa import JWE_RSA
from django.conf import settings


def encrypt_to_jwe(content):
    """Returns a string
    a serialized encryption from cryptojwt.jwe.jwe_rsa.JWE_RSA
    """
    if isinstance(content, dict):
        content = json.dumps(content).encode()
    elif isinstance(content, str):
        content = content.encode()

    if not isinstance(content, bytes):
        raise Exception('encrypt_to_jwe content must be a bytes object')

    RSA_KEY = settings.CAPTCHA_JWE_RSA_KEY_PATH
    JWE_ALG = settings.CAPTCHA_JWE_ALG
    JWE_ENC = settings.CAPTCHA_JWE_ENC

    priv_key = import_private_rsa_key_from_file(RSA_KEY)
    pub_key = priv_key.public_key()
    _rsa = JWE_RSA(content, alg=JWE_ALG, enc=JWE_ENC)
    jwe = _rsa.encrypt(pub_key)
    return jwe


def decrypt_from_jwe(jwe):
    RSA_KEY = settings.CAPTCHA_JWT_RSA_KEY_PATH
    JWE_ALG = settings.CAPTCHA_JWE_ALG
    JWE_ENC = settings.CAPTCHA_JWE_ENC

    priv_key = import_private_rsa_key_from_file(RSA_KEY)
    _decryptor = factory(jwe, alg=JWE_ALG, enc=JWE_ENC)
    _dkey = RSAKey(priv_key=priv_key)
    msg = _decryptor.decrypt(jwe, [_dkey])
    return msg



import ast
import re
import sys
from django import forms
from django.conf import settings

from . utils import _split_choices


def get_empty_form(form_class=forms.Form):
    class Dynamic(form_class):
        pass
    return Dynamic


def build_formset(choices, extra=0, required=False, prefix='form', data=None):
    """ Get formset
    """
    _regexp = '(?P<colname>[a-zA-Z0-9_]*)\((?P<coldict>[\{\}\.0-9a-zA-Z\'\"\:\;\_\,\s\-]*)\)'
    min_num = 0
    if required: min_num = 1
    eform = get_empty_form()
    for choice in choices:
        colname = choice # needed for simple CharField withoud attrs
        contenuto = re.search(_regexp, choice)
        field_dict = None
        if contenuto:
            coldict = contenuto.groupdict().get('coldict')
            colname = contenuto.groupdict()['colname']
            if coldict:
                field_dict = ast.literal_eval(coldict)
                field_type_name = field_dict['type']
                del field_dict['type']
                custom_field = getattr(sys.modules[__package__ + '.dynamic_fields'], field_type_name)(**field_dict) \
                               if hasattr(sys.modules[__package__ + '.dynamic_fields'], field_type_name) \
                               else getattr(sys.modules[__package__ + '.dynamic_fields'], 'CustomCharField')
                custom_widget = getattr(settings, 'CUSTOM_WIDGETS').get(field_type_name) \
                                if hasattr(settings, 'CUSTOM_WIDGETS') else None
                if custom_widget:
                    custom_field.widget = import_string(custom_widget)()

                if field_dict.get('choices'):
                    custom_field.choices += _split_choices(field_dict.get('choices'))
        else:
            custom_field = forms.CharField()
        eform.base_fields[colname] = custom_field
        eform.declared_fields[colname] = custom_field

    # Django formset
    fac =  forms.formset_factory(eform, extra=extra, min_num=min_num)
    if data: return fac(prefix=prefix, data=data)
    return fac(prefix=prefix)



import base64
import re

from django import forms
from django.conf import settings
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from . captcha import get_captcha
from . enc import encrypt, decrypt
from . formsets import build_formset
from . settings import FORMSET_TEMPLATE_NAMEID


FORMSET_TEMPLATE_NAMEID = getattr(settings, 'FORMSET_TEMPLATE_NAMEID',
                                  FORMSET_TEMPLATE_NAMEID)


class CaptchaWidget(forms.Widget):
    template_name = 'widgets/captcha.html'

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget as an HTML string."""
        context = self.get_context(name, value, attrs)
        # captcha = get_captcha(value)
        captcha = get_captcha(self.attrs['value'])
        # javascript functions don't allow "-" char
        unique_id = name.replace("-", "_")

        context['image_b64'] = base64.b64encode(captcha.read()).decode()
        context['widget']['value'] = ''
        context['widget']['hidden_field'] = self.attrs['hidden_field']
        context['widget']['unique_id'] = unique_id

        inline_code = mark_safe(
            '<script>'
            'function refresh_captcha_{unique_id}(){{'
                '$.get(location.href, function(data) {{'
                    '$("#{name}_img").attr("src",($(data).find("#{name}_img").attr("src")));'
                    '$("#{hidden_field}").val($(data).find("#{hidden_field}").val());'
                '}});'
            '}}'
            '</script>'.format(name=name,
                               unique_id=unique_id,
                               hidden_field=self.attrs['hidden_field'])
        )

        return self._render(self.template_name, context, renderer) + inline_code


class FormsetdWidget(forms.Widget):

    def __init__(self, *attrs, **kwargs):
        self.template = 'widgets/formset.html'
        self.readonly = False
        init_data_list = ['choices', 'data']
        field_required = kwargs.pop('field_required')
        self.prefix = kwargs.pop('prefix')
        for i in init_data_list:
            if i in kwargs.keys():
                setattr(self, i, kwargs.pop(i))

        if getattr(self, 'data'):
            # It MUST be a list!
            # self.data = [{'ccc': 'dfgdfgdfg', 'co': 'sono', 'col1': 'un', 'data': '2019-04-03'},
            #              {'ccc': '234234', 'co': 'son44o', 'col1': 'un333', 'data': '2016-04-03'}]
            self.formset = build_formset(choices=self.choices,
                                         required=field_required,
                                         prefix=self.prefix,
                                         data=self.data)
            self.formset.is_valid()
        else:
            # this initialized the formset as void
            self.formset = build_formset(choices=self.choices,
                                         required=field_required,
                                         extra=0,
                                         prefix=self.prefix)
        super().__init__(*attrs, **kwargs)

    def get_js_template(self):
        formset = build_formset(choices=self.choices,
                                prefix=self.prefix,
                                extra=1)
        res = re.sub('{}-0'.format(self.prefix),
                     '{}-{}'.format(self.prefix,
                                    FORMSET_TEMPLATE_NAMEID),
                     formset.forms[0].as_table())
        return mark_safe(res)

    def render(self, name='', value='', attrs=None, renderer=None):
        context_data = {'formset_id': self.prefix,
                        'template_generic_id': FORMSET_TEMPLATE_NAMEID,
                        'formset': self.formset,
                        'formset_template': self.get_js_template(),
                        'readonly': self.readonly,}
        html_output = render_to_string(self.template, context_data)
        return mark_safe(html_output)

    def make_readonly(self, attr='disabled'):
        self.readonly = True
        for form in self.formset:
            for generic_field in form:
                field = form.fields[generic_field.name]
                widget = field.widget
                widget.attrs[attr] = True
        return True



import copy

from django import forms
from django.conf import settings
from django.forms.fields import FileField
from django.template.defaultfilters import filesizeformat
from django.utils.module_loading import import_string


from . import dynamic_fields
from . dynamic_fields import CaptchaField, CaptchaHiddenField
from . utils import _split_choices_in_list_canc
from . widgets import FormsetdWidget


class BaseDynamicForm(forms.Form):
    def __init__(self,
                 fields_source=dynamic_fields,
                 initial_fields={},
                 final_fields={},
                 constructor_dict={},
                 custom_params={},
                 ignore_format_field_name=False,
                 *args,
                 **kwargs):
        """
        initial_fields:
        dyn_constructor_dict:
        example:

        """
        super().__init__(*args, **kwargs)

        # if initial fields are present
        self.fields = initial_fields or self.fields

        # Costruzione dinamica dei rimanenti fields del form
        if constructor_dict:
            constructor_dict = copy.deepcopy(constructor_dict)
            for key, value in constructor_dict.items():
                if ignore_format_field_name:
                    field_id = key
                else:
                    field_id = dynamic_fields.format_field_name(key)
                data_kwargs = {'label': key.title()}
                custom_field_name = value[0]
                custom_field_dict = value[1]
                pre_text = custom_field_dict.pop('pre_text')
                custom_field_values = value[2]
                # add custom attrs
                data_kwargs.update(custom_field_dict)
                custom_field = None
                if hasattr(fields_source, custom_field_name):
                    custom_field = getattr(fields_source,
                                           custom_field_name)(**data_kwargs)
                if custom_field:
                    custom_field.define_value(custom_field_values, **custom_params)
                    fields = custom_field.get_fields()
                    for field in fields:
                        name = getattr(field, 'name') if hasattr(field, 'name') else field_id
                        self.fields[name]= field
                        self.fields[name].pre_text = pre_text
                        if isinstance(field,
                                      dynamic_fields.CustomComplexTableField):
                            choices = _split_choices_in_list_canc(custom_field_values)
                            self.fields[name].widget = FormsetdWidget(choices=choices,
                                                                      data=kwargs.get('data', {}),
                                                                      field_required=field.required,
                                                                      prefix=name)
                        else:
                            custom_widget = getattr(settings, 'CUSTOM_WIDGETS', {}).get(field.__class__.__name__)
                            # Questo consente al widget di essere dichiarato anche come Tipo e non sempre come Stringa
                            if not custom_widget:
                                custom_widget = self.fields[name].widget
                            if isinstance(custom_widget, str):
                                self.fields[name].widget = import_string(custom_widget)()
                            else:
                                self.fields[name].widget = custom_widget
                            if hasattr(field, 'choices'):
                                self.fields[name].choices = getattr(field, 'choices')

        # if final fields are present
        self.fields.update(final_fields)

    def remove_not_compiled_fields(self):
        """
        Rimuove da un form compilato tutti i campi non compilati
        Viene usato nel metodo get_form
        """
        to_be_removed = []
        for field in self:
            form_field = self.fields[field.name]

            # remove not compiled single fields in formset
            if form_field.is_formset:
                formset = form_field.widget.formset
                for fform in formset:
                    to_be_removed_formset = []
                    for generic_field in fform:
                       if not generic_field.value():
                            to_be_removed_formset.append(generic_field.name)
                    for i in to_be_removed_formset:
                        del fform.fields[i]
                continue

            if not field.value():
                to_be_removed.append(field.name)
        for i in to_be_removed:
            del self.fields[i]

    def remove_datafields(self):
        """
        """
        to_be_removed = []
        for field in self.fields:
            if not isinstance(self.fields[field], FileField):
                # rimuove tutti i fields  che non sono File
                to_be_removed.append(field)
        for i in to_be_removed:
            del self.fields[i]

    def remove_files(self, allegati = None):
        """
        Rimuove tutti i FileFields
        viene usato in modifica_form se allegati già presenti
        """
        to_be_removed = []
        for field in self.fields:
            if isinstance(allegati, dict):
                # rimuove solo i fields allegati/files specificati
                if field in allegati:
                    to_be_removed.append(field)
            elif isinstance(self.fields[field], FileField):
                # rimuove tutti i fields  allegati/files
                to_be_removed.append(field)
        for i in to_be_removed:
            del self.fields[i]

    def clean(self, *args, **kwargs):
        cleaned_data = super().clean()

        self.data = copy.deepcopy(self.data)
        # CaPTCHA MUST BE ALWAYS RENEWED!
        for field_name, field_obj in self.fields.items():
            if type(field_obj) in (CaptchaField, CaptchaHiddenField):
                self.data[field_name] = self.fields[field_name].widget.attrs['value']
        # end CAPTCHA

        for fname in self.fields:
            field = self.fields[fname]
            # formset is empty or not valid
            if field.is_formset and not field.widget.formset.is_valid():
                errors = field.widget.formset.errors
                self.add_error(fname, errors)
                continue
            # other fields check
            if hasattr(field, 'parent'):
                field = getattr(field, 'parent')
                errors = field.raise_error(fname,
                                           cleaned_data,
                                           **kwargs)
            else:
                errors = field.raise_error(None,
                                           cleaned_data.get(fname),
                                           **kwargs)
            if errors:
                self.add_error(fname, errors)
                continue

    class Media:
        js = ('js/formset_js.js',)



import random
import string

from captcha.audio import AudioCaptcha
from captcha.image import ImageCaptcha

from django.conf import settings


def get_captcha(text=None):
    fonts = getattr(settings, 'CAPTCHA_FONTS', None)
    length = getattr(settings, 'CAPTCHA_LENGTH', 5)
    image = ImageCaptcha(fonts=fonts)

    text = text or ''.join([random.choice(string.ascii_letters+string.hexdigits) for i in range(length)])
    data = image.generate(text)
    return data

